#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <iostream>
#include "osc/OscOutboundPacketStream.h"
#include "osc/OscReceivedElements.h"
#include "ip/UdpSocket.h"
#include <errno.h>
#include "bergenSound.h"
#include "bergenServer.h"
#include "bergenListener.h"

#define NUM_TRIES 2
#define WAIT_TICKS 50
#define MAX_RECEIVE_BUFFER 4096
#define MAX_SEND_BUFFER 1024

char * bergenVersion = "bergen client library, Version 0.4.1 - Compiled " __DATE__;

struct _sound
	{
	bergenSound * sound;
	struct _sound * next;
	};

// The constructor will make a UDP network connection to the snerd server     
// program, sending it a "ping" message to verify the connection. The argument
// host, if given, is the name or IP address of the machine which snerd is    
// running on; if it is not given, the value of the environment variable      
// BERGEN_SERVER is used; if this variable is not set, the connection is made 
// to the local machine. The argument port is the port number that the server 
// is listening to; if it is not given (or is 0), the value of the environment 
// variable BERGEN_PORT is used; if this variable is not set, the default port
// of 5900 is used.
bergenServer::bergenServer(char *server,int port)
	{
	soundList_ = NULL;
	path_ = NULL;
	listenerClient_ = NULL;

	if (server)
		serverName_ = strdup(server);
	else if (getenv("BERGEN_SERVER"))
		serverName_ = strdup(getenv("BERGEN_SERVER"));
	else
		serverName_ = "localhost";
	if (port)
		serverPort_ = port;
	else if (getenv("BERGEN_PORT"))
		serverPort_ = atoi(getenv("BERGEN_PORT"));
	else
		serverPort_ = 5900;
	transmitSocket = new UdpTransmitSocket( IpEndpointName( serverName_, serverPort_ ) );
	transmitSocket->NonBlocking();
	receiveSocket = new UdpReceiveSocket( IpEndpointName( IpEndpointName::ANY_ADDRESS, serverPort_+1 ) );
	receiveSocket->NonBlocking();
	connect();
	listenerClient_ = new bergenListener(this);
	if (getenv("BERGEN_PATH"))
		addPath(getenv("BERGEN_PATH"));
	}


// The destructor will automatically delete all sounds which are in its list 
// of bergenSound objects (maintained by addSound() and removeSound).
bergenServer::~bergenServer(void)
	{
	struct _sound *s,*next=NULL;
	for (s=soundList_; s; s = next)
		{
		next = s->next;
		delete s->sound;
		}
	if (path_)
		free(path_);
	if (listenerClient_)
	 	free(listenerClient_);
	}


// Tries to establish a connection to the snerd server. Error messages if no
// reply from the snerd server or no socket is given in the bergenServer
void bergenServer::connect(void)
	{
	int tries, wait;
	if (!transmitSocket || !receiveSocket)
		{
		fprintf(stderr,"ERROR (bergenServer::connect()): don't have a socket to use for connection\n");
		return;
		}

	char *data = new char[MAX_RECEIVE_BUFFER];
	IpEndpointName remoteEndpoint;

	for (tries=0; tries < NUM_TRIES; tries++)
		{
		char buffer[1024];
		osc::OutboundPacketStream p( buffer, MAX_SEND_BUFFER );
		p << osc::BeginMessage( "server" ) << "ping" << osc::EndMessage;
		transmitSocket->Send( p.Data(), p.Size() );
		for (wait=0; wait < WAIT_TICKS; wait++)
			{
			int size = receiveSocket->ReceiveFrom( remoteEndpoint, data, MAX_RECEIVE_BUFFER );
			if( size > 0 )
				{
        			osc::ReceivedPacket p( data, size );
				osc::ReceivedMessage* msgPtr;
				if (p.IsBundle())
					{
					osc::ReceivedBundle b(p);
					osc::ReceivedBundle::const_iterator i = b.ElementsBegin();
					msgPtr = new osc::ReceivedMessage(*i);
					}
				else
					msgPtr = new osc::ReceivedMessage(p);
				if( strcmp( msgPtr->AddressPattern(), "client" ) == 0 )
					{
					osc::ReceivedMessage::const_iterator arg = msgPtr->ArgumentsBegin();
					const char *str = arg->AsString();
					if (!strcasecmp(str,"pong"))
						break;
					}
				delete msgPtr;
				}
			usleep(10000);
			}
		if (wait < WAIT_TICKS)
			break;
		}
	if (tries == NUM_TRIES)
		{
		fprintf(stderr,"WARNING: bergenServer failed to connect to server (%s)\n",serverName_);
		delete transmitSocket;
		transmitSocket = NULL;
		delete receiveSocket;
		receiveSocket = NULL;
		}
	delete [] data;
	}


// Re-initializes the connection to the server program, and re-creates all the
// sound objects (that it manages) on the server. This can be used to reset
// things if snerd crashes & is restarted, or if it is started after the
// bergenServer object is created. 
void bergenServer::reset(void)
	{
	struct _sound *s;
	for (s=soundList_; s; s = s->next)
		s->sound->killRemote();
	if (!transmitSocket)
		{
		transmitSocket = new UdpTransmitSocket( IpEndpointName( serverName_, serverPort_ ) );
		transmitSocket->NonBlocking();
		receiveSocket = new UdpReceiveSocket(IpEndpointName( IpEndpointName::ANY_ADDRESS, serverPort_+1 ));
		receiveSocket->NonBlocking();
		}
	connect();
	if (transmitSocket)
		for (s=soundList_; s; s = s->next)
			s->sound->createRemote();
	}


// If a message was received over the socket, the message is send to 
// splitMessage for getting messages isolated. The isolated massages are
// parsed by sending them to parseMessage if they are no sound related ones.
// Sound messages are forwarded to the according sound object
void bergenServer::processMessages(void)
	{
	if (receiveSocket)
		{
		char *data = new char[MAX_RECEIVE_BUFFER];
		IpEndpointName remoteEndpoint;
		int size = receiveSocket->ReceiveFrom( remoteEndpoint, data, MAX_RECEIVE_BUFFER );
		if( size > 0 )
			{
			osc::ReceivedPacket p( data, size );
			osc::ReceivedMessage* msgPtr;
			if (p.IsBundle())
				{
				osc::ReceivedBundle b(p);
				osc::ReceivedBundle::const_iterator i = b.ElementsBegin();
				msgPtr = new osc::ReceivedMessage(*i);
				}
			else
				msgPtr = new osc::ReceivedMessage(p);
			if( strcmp( msgPtr->AddressPattern(), "object" ) == 0 )
				{			
				osc::ReceivedMessage::const_iterator arg = msgPtr->ArgumentsBegin();
				int handle = (arg++)->AsInt32();
				bergenSound* sound = findHandle(handle);
				if (sound)
					sound->message(arg);
				else
					printf("WARNING: bergenServer failed to find sound with handle %d\n",handle);
				}
			delete msgPtr;
			}
		delete [] data;
		}
	}


// Adds a sound to the server object's list. This is used internally, by the
// bergenSound class, and should not be called by applications.
// The list consists of the references to the sound objects. The sound object
// data is managed by snerd 
void bergenServer::addSound(bergenSound * sound)
	{
	struct _sound *s = new struct _sound;
	s->sound = sound;
	s->next = soundList_;
	soundList_ = s;
	}


// Finds a sound given the handle
bergenSound* bergenServer::findHandle(int handle)
	{
	struct _sound *s;
	if (!soundList_)
		return NULL;
	for (s=soundList_; s; s = s->next)
		{
		if (s->sound->handle() == handle)
			return s->sound;
		}
	return NULL;
	}


// Removes a sound from the server object's list. This is used internally, by 
// the bergenSound class, and should not be called by applications 
void bergenServer::removeSound(bergenSound * sound)
	{
	struct _sound *s, *removed=NULL;
	if (!soundList_)
		return;
	if (soundList_->sound == sound)
		{
		removed = soundList_;
		soundList_ = soundList_->next;
		}
	else
		for (s=soundList_; s->next; s = s->next)
			{
			if (s->next->sound == sound)
				{
				removed = s->next;
				s->next = s->next->next;
				break;
				}
			}
	if (removed)
		delete removed;
	}


// Sets the default directory path for sound objects. bergenSample objects will
// use this path if their sample file name does not include a path.
// The path is used by snerd to look up the sound files which are to be
// stored at sound server's location
void bergenServer::addPath(char *path)
	{
	if (path_)
		free(path_);
	if (path)
		{
		path_ = strdup(path);
		char buffer[MAX_SEND_BUFFER];
		osc::OutboundPacketStream p( buffer, MAX_SEND_BUFFER );
		p << osc::BeginMessage( "server" ) << "path" << path_ << osc::EndMessage;
		sendMessage( p );
		}
	else
		path_ = NULL;
	}


// Returns the default directory path set by setPath().
char * bergenServer::path(void)
	{
	return path_;
	}


// Sends the given text string to the server program. This is mostly for
// internal use, by the bergenSound classes. The snerd analyses the message
// and reacts on it.
void bergenServer::sendMessage(const osc::OutboundPacketStream& p)
	{
	if (transmitSocket)
		transmitSocket->Send( p.Data(), p.Size() );	
	}


// Wait for the next message from the server
int bergenServer::receiveHandle(void)
	{
	if (receiveSocket)
		{
		int wait;
		int handle;
		char *data = new char[MAX_RECEIVE_BUFFER];
		IpEndpointName remoteEndpoint;
		for (wait=0; wait < WAIT_TICKS; wait++)
			{
			int size = receiveSocket->ReceiveFrom( remoteEndpoint, data, MAX_RECEIVE_BUFFER );
			if( size > 0 )
				{
				osc::ReceivedPacket p( data, size );
				osc::ReceivedMessage* msgPtr;
				if (p.IsBundle())
					{
					osc::ReceivedBundle b(p);
					osc::ReceivedBundle::const_iterator i = b.ElementsBegin();
					msgPtr = new osc::ReceivedMessage(*i);
					}
				else
					msgPtr = new osc::ReceivedMessage(p);
				if( strcmp( msgPtr->AddressPattern(), "client" ) == 0 )
					{			
					osc::ReceivedMessage::const_iterator arg = msgPtr->ArgumentsBegin();
					if (arg->IsString())
						{
						const char* str = (arg++)->AsString();
						if (!strcasecmp(str,"handle"))
							{
							handle = arg->AsInt32();
							delete[] data;
							return handle;
							}
						}
					}
				delete msgPtr;	
				}
			usleep(10000);
			}
		fprintf(stderr,"WARNING: bergenServer failed to receive a handle from (%s)\n",serverName_);
		delete[] data;
		}
	return 0;
	}
